package org.processmining.glueminer.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XExtendedEvent;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.glueminer.models.ActivitiesOccurrencesMemorizer;
import org.processmining.glueminer.models.Behaviors;
import org.processmining.glueminer.models.DirectlyFollowsMatrix;
import org.processmining.glueminer.models.IndirectlyFollowsMatrix;
import org.processmining.glueminer.models.OrderedActivity;
import org.processmining.glueminer.models.PlaceUnderfednessStorage;
import org.processmining.glueminer.models.PlacesMemorizer;
import org.processmining.glueminer.parameters.GlueMinerParameters;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.etconformance.ETCPlugin;
import org.processmining.plugins.etconformance.ETCResults;
import org.processmining.plugins.etconformance.ETCSettings;

public class GlueMiner {
	Double placeFitnessLevel;
	Double placeUnderfednessLevel;
	Double netFitnessLevel;
	Double relativeThresholdOnSourceAndEndActivities;
	Double relativeThresholdOnPlacesRegardingLogSize;
	Integer minSupportForPlaces;
	Integer maxSourceOrTargetTransitions;
	Integer minAcceptedBehaviorCount;
	
	PluginContext context;
	XLog log;
	Behaviors behaviors;
	XEventClassifier classifier;
	XEventClass dummyEvClass;
	XLogInfo summary;
	Collection<XEventClass> logClasses;
	Petrinet net;
	TransEvClassMapping mapping;
	Map<String, Transition> activitiesTransitions;
	Map<String, Integer> startActivities;
	Map<String, Integer> endActivities;
	Place sourcePlace;
	Place sinkPlace;
	Marking initialMarking;
	Marking finalMarking;
	DirectlyFollowsMatrix directlyFollowsMatrix;
	IndirectlyFollowsMatrix indirectlyFollowsMatrix;
	ActivitiesOccurrencesMemorizer activitiesOccurrences;
	PlacesMemorizer placesMemorizer;
	Set<String> possibleLeftActivitiesForPlacesAlgorithm;
	Set<String> possibleRightActivitiesForPlacesAlgorithm;
	Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheLeftSide;
	Map<OrderedActivity, Set<OrderedActivity>> nonImportantRelationshipsOnTheRightSide;
	Map<OrderedActivity, Set<OrderedActivity>> possibleActivitiesOnTheRightMap;
	
	
	TokenReplayer replayer;
	PlaceAdditionTriager placeAdditionTriager;
	PlaceSearchRecursion placeSearchRecursion;
	
	public static Integer removedUnfitPlaces = 0;
	
	public GlueMiner(PluginContext context, XLog log, GlueMinerParameters parameters) {
		this.buildObject(context, log, parameters.placeFitnessLevel, parameters.maxUnderfednessLevel, parameters.netFitnessLevel, parameters.relThresholdOnStartAndEndActivities, parameters.relThresholdOnPlacesLogSize, parameters.maxSourceOrTargetTransitions, parameters.minAcceptedBehaviorCount);
	}
	
	public GlueMiner(PluginContext context, XLog log, Double wantedFitnessLevel, Double placesUnderfednessLevel, Double netFitnessLevel, Double relativeThresholdOnSourceAndEndActivities, Double relativeThresholdOnPlacesRegardingLogSize, Integer maxSourceOrTargetTransitions, Integer minAcceptedBehaviorCount) {
		this.buildObject(context, log, wantedFitnessLevel, placesUnderfednessLevel, netFitnessLevel, relativeThresholdOnSourceAndEndActivities, relativeThresholdOnPlacesRegardingLogSize, maxSourceOrTargetTransitions, minAcceptedBehaviorCount);
	}
	
	public void buildObject(PluginContext context, XLog log, Double wantedFitnessLevel, Double placesUnderfednessLevel, Double netFitnessLevel, Double relativeThresholdOnSourceAndEndActivities, Double relativeThresholdOnPlacesRegardingLogSize, Integer maxSourceOrTargetTransitions, Integer minAcceptedBehaviorCount) {
		GlueMiner.removedUnfitPlaces = 0;
		this.context = context;
		this.relativeThresholdOnSourceAndEndActivities = relativeThresholdOnSourceAndEndActivities;
		this.relativeThresholdOnPlacesRegardingLogSize = relativeThresholdOnPlacesRegardingLogSize;
		this.placeFitnessLevel = wantedFitnessLevel;
		this.placeUnderfednessLevel = placesUnderfednessLevel;
		this.netFitnessLevel = netFitnessLevel;
		this.maxSourceOrTargetTransitions = maxSourceOrTargetTransitions;
		this.minAcceptedBehaviorCount = minAcceptedBehaviorCount;
		
		// debugging purposes
		this.log = log;
		if (!this.log.getAttributes().containsKey("concept:name")) {
			this.log.getAttributes().put("concept:name", new XAttributeLiteralImpl("concept:name", "emptyname"));
		}
		
		this.minSupportForPlaces = (int)(this.log.size() * this.relativeThresholdOnPlacesRegardingLogSize);
		this.cleanNonCompleteTransitions();
		this.classifier = XLogInfoImpl.NAME_CLASSIFIER;
		this.dummyEvClass = new XEventClass("DUMMYDUMMYDUMMY", 999999);
		this.mapping = new TransEvClassMapping(this.classifier, this.dummyEvClass);
		
		this.summary = XLogInfoFactory.createLogInfo(log, this.classifier);
		

		this.logClasses = this.summary.getEventClasses().getClasses();
		this.buildInitialPetrinet();
		this.buildStartActivities();
		this.buildEndActivities();
		this.connectStartAndEndEventToStartAndEndActivities();
		
		this.behaviors = new Behaviors(activitiesTransitions);
		this.behaviors.initialize(log);
		this.behaviors.cleanBehaviors(this.minAcceptedBehaviorCount);
		//this.behaviors.printResume();
		
		//System.out.println("placeFitnessLevel = "+this.placeFitnessLevel);
	}
	
	public Petrinet getNet() {
		return this.net;
	}
	
	public Marking getInitialMarking() {
		return this.initialMarking;
	}
	
	public Marking getFinalMarking() {
		return this.finalMarking;
	}
	
	public XEventClassifier getClassifier() {
		return this.classifier;
	}
	
	public TransEvClassMapping getMapping() {
		return this.mapping;
	}
	
	public void cleanNonCompleteTransitions() {
		int i = 0;
		int j = 0;
		while (i < this.log.size()) {
			XTrace tr = this.log.get(i);
			j = 0;
			while (j < tr.size()) {
				XExtendedEvent eve = XExtendedEvent.wrap(tr.get(j));
				if (!(eve.getTransition() == null || eve.getTransition().isEmpty() || eve.getTransition().toLowerCase().equals("complete"))) {
					tr.remove(j);
					continue;
				}
				j++;
			}
			i++;
		}
	}
	
	public void buildStartActivities() {
		this.startActivities = new HashMap<String, Integer>();
		
		for (XTrace tr : log) {
			if (tr.size() > 0) {
				String aName = XExtendedEvent.wrap(tr.get(0)).getName();
				if (!this.startActivities.containsKey(aName)) {
					this.startActivities.put(aName, 0);
				}
				this.startActivities.put(aName, this.startActivities.get(aName) + 1);
			}
		}
		
		List<String> startActivitiesKeys = new ArrayList<String>(this.startActivities.keySet());
		Integer maximumCountForStartActivities = 0;
		
		for (String startActivity : startActivitiesKeys) {
			if (this.startActivities.get(startActivity) > maximumCountForStartActivities) {
				maximumCountForStartActivities = this.startActivities.get(startActivity);
			}
			if (this.startActivities.get(startActivity) < log.size() * (1.0 - this.placeFitnessLevel)) {
				System.out.println("(1) removing start activity: "+startActivity+" with count: "+this.startActivities.get(startActivity));
				this.startActivities.remove(startActivity);
			}
		}
		
		startActivitiesKeys = new ArrayList<String>(this.startActivities.keySet());
		
		for (String startActivity : startActivitiesKeys) {
			if (this.startActivities.get(startActivity) < maximumCountForStartActivities * relativeThresholdOnSourceAndEndActivities) {
				System.out.println("(2) removing start activity: "+startActivity+" with count: "+this.startActivities.get(startActivity));
				this.startActivities.remove(startActivity);
			}
		}
	}
	
	public void buildEndActivities() {
		this.endActivities = new HashMap<String, Integer>();
		
		for (XTrace tr : log) {
			if (tr.size() > 0) {
				String aName = XExtendedEvent.wrap(tr.get(tr.size()-1)).getName();
				if (!this.endActivities.containsKey(aName)) {
					this.endActivities.put(aName, 0);
				}
				this.endActivities.put(aName, this.endActivities.get(aName) + 1);
			}
		}
		
		List<String> endActivitiesKeys = new ArrayList<String>(this.endActivities.keySet());
		Integer maximumCountForEndActivities = 0;
		
		for (String endActivity : endActivitiesKeys) {
			if (this.endActivities.get(endActivity) > maximumCountForEndActivities) {
				maximumCountForEndActivities = this.endActivities.get(endActivity);
			}
			if (this.endActivities.get(endActivity) < log.size() * (1.0 - this.placeFitnessLevel)) {
				System.out.println("(1) removing end activity: "+endActivity+" with count: "+this.endActivities.get(endActivity));
				this.endActivities.remove(endActivity);
			}
		}
		
		endActivitiesKeys = new ArrayList<String>(this.endActivities.keySet());
		
		for (String endActivity : endActivitiesKeys) {
			if (this.endActivities.get(endActivity) < maximumCountForEndActivities * relativeThresholdOnSourceAndEndActivities) {
				System.out.println("(2) removing end activity: "+endActivity+" with count: "+this.endActivities.get(endActivity));
				this.endActivities.remove(endActivity);
			}
		}
	}
	
	
	public void buildInitialPetrinet() {
		this.activitiesTransitions = new HashMap<String, Transition>();
		this.possibleLeftActivitiesForPlacesAlgorithm = new HashSet<String>();
		this.possibleRightActivitiesForPlacesAlgorithm = new HashSet<String>();
		
		this.net = new PetrinetImpl("");
		for (XEventClass logClass : this.logClasses) {
			String activity = logClass.getId();
			Transition transition = this.net.addTransition(activity);
			this.activitiesTransitions.put(activity, transition);
			this.possibleLeftActivitiesForPlacesAlgorithm.add(activity);
			this.possibleRightActivitiesForPlacesAlgorithm.add(activity);
			
			this.mapping.put(transition, logClass);
		}
	}
	
	public void connectStartAndEndEventToStartAndEndActivities() {
		this.sourcePlace = net.addPlace("source");
		this.sinkPlace = net.addPlace("sink");
		this.initialMarking = new Marking();
		this.initialMarking.add(this.sourcePlace);
		this.finalMarking = new Marking();
		this.finalMarking.add(this.sinkPlace);
		
		List<String> startActivitiesKeys = new ArrayList<String>(this.startActivities.keySet());
		List<String> endActivitiesKeys = new ArrayList<String>(this.endActivities.keySet());
		
		for (String startActivity : startActivitiesKeys) {
			if (endActivities.containsKey(startActivity) && endActivities.get(startActivity) > startActivities.get(startActivity)) {
				System.out.println("(3) removing start activity "+startActivity);
				this.startActivities.remove(startActivity);
			}
		}
		
		for (String endActivity : endActivitiesKeys) {
			if (startActivities.containsKey(endActivity) && startActivities.get(endActivity) > endActivities.get(endActivity)) {
				System.out.println("(3) removing end activity "+endActivity);
				this.endActivities.remove(endActivity);
			}
		}
		
		//System.out.println("this.startActivities = "+this.startActivities);
		//System.out.println("this.endActivities = "+this.endActivities);
		
		for (String activity : this.startActivities.keySet()) {
			net.addArc(this.sourcePlace, this.activitiesTransitions.get(activity));
		}
		for (String activity : this.endActivities.keySet()) {
			net.addArc(this.activitiesTransitions.get(activity), this.sinkPlace);
		}
		
		this.possibleRightActivitiesForPlacesAlgorithm.removeAll(this.startActivities.keySet());
		this.possibleLeftActivitiesForPlacesAlgorithm.removeAll(this.endActivities.keySet());
	}
	
	public void mineEntireProcess() {
		Double allPossiblePlacesEvaluatedByGreedyAlgorithm = Math.pow(2.0, 2.0 * (this.mapping.keySet().size() + 1));
		System.out.println("number of places that would be evaluated by greedy algorithm: "+allPossiblePlacesEvaluatedByGreedyAlgorithm);
		
		calculateDirectlyAndIndirectlyFollows();
		applyTokenReplaying();
		removeUnfitPlaces();
		initializePlaceAdditionTriager();
		initializePlaceSearchRecursion();
		startRecursion();
		this.placeSearchRecursion.printReview();
		applyTokenReplaying();
		removeUnfitPlaces();
		applyTokenReplaying();
		//System.out.println("this.replayer.traceFitness");
		//System.out.println(this.replayer.traceFitness);
		if (this.replayer.traceFitness < this.netFitnessLevel) {
			powerCleanTheNetToReachFitnessLevel();
		}
		System.out.println("removed unfit places at end of algorithm "+GlueMiner.removedUnfitPlaces);
		
		//printReplayerResume();
	}
	
	public void powerCleanTheNetToReachFitnessLevel() {
		List<PlaceUnderfednessStorage> underfednessStorage = new ArrayList<PlaceUnderfednessStorage>(this.replayer.getPlaceUnderfednessStorage());
		int i = 0;
		while (i < underfednessStorage.size()) {
			Place p = underfednessStorage.get(i).place;
			System.out.println("powercleaning removing place "+p.getLabel()+" (fitness: "+this.replayer.placeFitness.get(p)+", underfedness: "+this.replayer.placeUnderfedness.get(p)+")");
			
			GlueMiner.removedUnfitPlaces++;
			
			removePlaceFromNet(p);
			applyTokenReplaying();
			if (this.replayer.traceFitness >= this.netFitnessLevel) {
				break;
			}
			i++;
		}
	}
	
	public void initializePlaceAdditionTriager() {
		this.placeAdditionTriager = new PlaceAdditionTriager(this.behaviors, this.net, this.initialMarking, this.finalMarking, this.placesMemorizer, this.activitiesTransitions);
	}
	
	public void initializePlaceSearchRecursion() {
		this.placeSearchRecursion = new PlaceSearchRecursion(this.placeFitnessLevel, this.placeUnderfednessLevel, this.placeAdditionTriager, this.nonImportantRelationshipsOnTheLeftSide, this.nonImportantRelationshipsOnTheRightSide, this.possibleActivitiesOnTheRightMap, this.maxSourceOrTargetTransitions, this.minSupportForPlaces);
	}
	
	public void startRecursion() {
		this.placeSearchRecursion.startRecursion();
	}
	
	public void applyTokenReplaying() {
		this.replayer = new TokenReplayer(this.behaviors, this.net, this.initialMarking, this.finalMarking);
		this.replayer.doReplay();
		//this.replayer.printResume();
	}
	
	public void printReplayerResume() {
		this.replayer.printResume();
	}
	
	public void removeUnfitPlaces() {
		List<Place> placesToConsider = new ArrayList<Place>(this.replayer.placeFitness.keySet());
		
		for (Place place : placesToConsider) {
			if (place.getLabel() != "source" && place.getLabel() != "sink") {
				boolean condition = false;
				if (this.replayer.placeFitness.get(place) < this.placeFitnessLevel) {
					condition = true;
				}
				if (this.replayer.placeUnderfedness.get(place) > this.placeUnderfednessLevel) {
					condition = true;
				}
				if (condition) {
					GlueMiner.removedUnfitPlaces++;
					removePlaceFromNet(place);
				}
			}
		}
	}
	
	public void removePlaceFromNet(Place place) {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = net.getInEdges(place);
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = net.getOutEdges(place);
		
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : inEdges) {
			net.removeEdge(edge);
		}
		
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : outEdges) {
			net.removeEdge(edge);
		}
		
		net.removePlace(place);
	}
	
	public void calculateDirectlyAndIndirectlyFollows() {
		this.directlyFollowsMatrix = new DirectlyFollowsMatrix();
		this.indirectlyFollowsMatrix = new IndirectlyFollowsMatrix();
		this.placesMemorizer = new PlacesMemorizer();
		this.activitiesOccurrences = new ActivitiesOccurrencesMemorizer();
		
		for (XTrace tr : this.log) {
			int i = 0;
			while (i < tr.size()) {
				String activity1 = XExtendedEvent.wrap(tr.get(i)).getName();
				this.activitiesOccurrences.addActivityOccurrence(activity1, i);
				int j = i+1;
				while (j < tr.size()) {
					String activity2 = XExtendedEvent.wrap(tr.get(j)).getName();
					if (j == (i+1)) {
						this.directlyFollowsMatrix.addRelationshipBetweenActivities(activity1, activity2);
					}
					this.indirectlyFollowsMatrix.addRelationshipBetweenActivities(activity1, activity2);
					j++;
				}
				i++;
			}
		}
		
		this.activitiesOccurrences.calculateAverage();
		this.directlyFollowsMatrix.removeActivitiesBelowSupport(this.minSupportForPlaces);
		this.directlyFollowsMatrix.clean(this.placeFitnessLevel);
		
		this.indirectlyFollowsMatrix.clean(this.placeFitnessLevel);
		this.indirectlyFollowsMatrix.calculateImportantRelationships(this.activitiesOccurrences.getActivitiesCount(), this.placeFitnessLevel);
		this.indirectlyFollowsMatrix.calculateReallyImportantRelationships();
		//this.indirectlyFollowsMatrix.printResume();
		
		Integer relationCount = this.directlyFollowsMatrix.addEvidentRelations(this.placesMemorizer, this.possibleLeftActivitiesForPlacesAlgorithm, this.possibleRightActivitiesForPlacesAlgorithm, this.minSupportForPlaces, this.activitiesOccurrences.getActivitiesCount());
		this.possibleLeftActivitiesForPlacesAlgorithm.removeAll(this.placesMemorizer.getAllInputActivitiesInPlaces());
		this.possibleRightActivitiesForPlacesAlgorithm.removeAll(this.placesMemorizer.getAllOutputActivitiesInPlaces());
		relationCount += this.indirectlyFollowsMatrix.addEvidentRelations(this.placesMemorizer, this.possibleLeftActivitiesForPlacesAlgorithm, this.possibleRightActivitiesForPlacesAlgorithm, this.minSupportForPlaces, this.activitiesOccurrences.getActivitiesCount());
		this.possibleLeftActivitiesForPlacesAlgorithm.removeAll(this.placesMemorizer.getAllInputActivitiesInPlaces());
		this.possibleRightActivitiesForPlacesAlgorithm.removeAll(this.placesMemorizer.getAllOutputActivitiesInPlaces());
		
		System.out.println("evident relations added before recursion with direct/indirectly follows matrix calculation: "+relationCount);
		
		this.placesMemorizer.addToPetrinet(net, activitiesTransitions);
		
		this.nonImportantRelationshipsOnTheLeftSide = this.indirectlyFollowsMatrix.getNonImportantRelationships(this.possibleLeftActivitiesForPlacesAlgorithm, 1, this.activitiesOccurrences);
		this.nonImportantRelationshipsOnTheRightSide = this.indirectlyFollowsMatrix.getNonImportantRelationships(this.possibleRightActivitiesForPlacesAlgorithm, -1, this.activitiesOccurrences);
		this.possibleActivitiesOnTheRightMap = this.indirectlyFollowsMatrix.getPossibleActivitiesOnTheRight(this.activitiesOccurrences);
	}
	
	public ETCResults getEtcResults() {
		ETCResults res = new ETCResults();
		ETCSettings sett = new ETCSettings(res);
		ETCPlugin etcPlugin = new ETCPlugin();
		Object[] etcResults = etcPlugin.doETC(context, log, net, initialMarking, mapping, res);
		ETCResults resNew = (ETCResults) etcResults[0];
		
		return resNew;
	}
	
	public static void main(String[] args) {
		/*PluginContextFactory contextFactory = new PluginContextFactory();
		PluginContext context = contextFactory.getContext();
		
		XLog log = ReadEventLog.readSingleLog("c:\\running-example.xes");
		GlueMiner miner = new GlueMiner(context, log, GlueMinerConstants.defaultPlaceFitnessLevel, GlueMinerConstants.defaultMaxUnderfednessLevel, GlueMinerConstants.defaultMinNetFitness, GlueMinerConstants.defaultRelativeThresholdOnStartAndEndActivities, GlueMinerConstants.defaultRelativeThresholdOnPlacesRegardingLogSize, GlueMinerConstants.defaultMaxSourceOrTargetTransitions, GlueMinerConstants.defaultMinAcceptedBehaviorCount);
		
		miner.mineEntireProcess();
		
		ETCResults conformanceResults = miner.getEtcResults();
		long nonFitTraces = conformanceResults.getnNonFitTraces();
		double fitness = 0.0;
		if (log.size() > 0) {
			fitness = 1.0 - (double)nonFitTraces/(double)log.size();
		}
		
		System.out.println("ETC non fit traces: "+nonFitTraces+" out of "+log.size()+" (fitness: "+fitness+")");
		System.out.println("ETC precision: "+conformanceResults.getEtcp());
		ExportPetriNet.exportPetrinetToPNMLorEPNMLFile(miner.getNet(), miner.getFinalMarking(), "C:\\users\\berti\\running-example.pnml");
		
		System.out.println("finished");*/
	}
}
