package org.processmining.glueminer.models;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ActivitiesOccurrencesMemorizer {
	public Map<String, List<Integer>> activitiesOccurrences;
	public Map<String, Double> averageActivityOccurrence;
	public Map<String, Integer> activityOverallCount;
	public Map<String, OrderedActivity> averageOrderedActivities;
	
	public ActivitiesOccurrencesMemorizer() {
		this.activitiesOccurrences = new HashMap<String, List<Integer>>();
		this.averageActivityOccurrence = new HashMap<String, Double>();
		this.activityOverallCount = new HashMap<String, Integer>();
		this.averageOrderedActivities = new HashMap<String, OrderedActivity>();
	}
	
	public void addActivityOccurrence(String activity, Integer position) {
		if (!activitiesOccurrences.containsKey(activity)) {
			activitiesOccurrences.put(activity, new ArrayList<Integer>());
		}
		activitiesOccurrences.get(activity).add(position);
	}
	
	public Double calculateListAverage(List<Integer> values) {
		double sum = 0.0;
		double count = values.size();
		
		for (Integer value : values) {
			sum += value;
		}
		
		if (count > 0.5) {
			return sum / count;
		}
		return 0.0;
	}
	
	public void calculateAverage() {
		for (String activity : activitiesOccurrences.keySet()) {
			double listAverage = calculateListAverage(activitiesOccurrences.get(activity));
			averageActivityOccurrence.put(activity, listAverage);
			averageOrderedActivities.put(activity, new OrderedActivity(activity, listAverage, activitiesOccurrences.get(activity).size()));
			activityOverallCount.put(activity, activitiesOccurrences.get(activity).size());
		}
	}
	
	public Map<String, Double> getAverage() {
		return this.averageActivityOccurrence;
	}
	
	public Map<String, Integer> getActivitiesCount() {
		return this.activityOverallCount;
	}
}
