package org.processmining.glueminer.models;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.extension.std.XExtendedEvent;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class Behaviors {
	public Map<String, Transition> activitiesTransitions;
	public Map<String, Behavior> behaviors;
	
	public Behaviors(Map<String, Transition> activitiesTransitions) {
		this.activitiesTransitions = activitiesTransitions;
		this.behaviors = new HashMap<String, Behavior>();
	}
	
	public List<Behavior> getBehaviorList() {
		return new ArrayList<Behavior>(this.behaviors.values());
	}
	
	public void addBehavior(String variant) {
		if (!behaviors.containsKey(variant)) {
			behaviors.put(variant, new Behavior(variant, activitiesTransitions));
		}
		behaviors.get(variant).increase();
	}
	
	public void initialize(XLog log) {
		for (XTrace tr : log) {
			StringBuilder variant = new StringBuilder();
			
			for (XEvent ev : tr) {
				XExtendedEvent eve = XExtendedEvent.wrap(ev);
				if (variant.length() > 0) {
					variant.append(",");
				}
				variant.append(eve.getName());
			}
			
			this.addBehavior(variant.toString());
		}
		this.buildEnabledTransitions();
	}
	
	public void cleanBehaviors(Integer threshold) {
		List<String> variants = new ArrayList<String>(this.behaviors.keySet());
		
		// memorize max count
		int maxVariantCount = 0;
		for (String variant : variants) {
			if (behaviors.get(variant).getVariantCount() > maxVariantCount) {
				maxVariantCount = behaviors.get(variant).getVariantCount();
			}
		}
		
		for (String variant : variants) {
			if (behaviors.get(variant).getVariantCount() < Math.min(maxVariantCount, threshold)) {
				behaviors.remove(variant);
			}
		}
	}
	
	public void buildEnabledTransitions() {
		for (Behavior behavior : behaviors.values()) {
			behavior.buildEnabledTransitions();
		}
	}
	
	public void printResume() {
		for (Behavior behavior : behaviors.values()) {
			behavior.printResume();
		}
	}
}
