package org.processmining.glueminer.plugins;

import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.wizard.ListWizard;
import org.processmining.framework.util.ui.wizard.ProMWizardDisplay;
import org.processmining.framework.util.ui.wizard.ProMWizardStep;
import org.processmining.glueminer.algorithms.GlueMiner;
import org.processmining.glueminer.dialogs.GlueMinerWizardStep;
import org.processmining.glueminer.parameters.GlueMinerParameters;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.etconformance.ETCResults;

@Plugin(name = "Glue Miner (Process Discovery)", returnLabels = { "Petrinet", "Initial marking", "Final marking", "TransEvClassMapping" }, returnTypes = { Petrinet.class, Marking.class, Marking.class, TransEvClassMapping.class }, parameterLabels = {
"XLog" }, help = "Apply Glue Miner to an Event Log", userAccessible = true)
public class GlueMinerPlugin {
	@UITopiaVariant(author = "Alessandro Berti", email = "a.berti@pads.rwth-aachen.de", affiliation = "PADS")
	@PluginVariant(requiredParameterLabels = { 0 })
	public Object[] getPetrinet(UIPluginContext context, XLog log) {
		Object[] ret = null;
		
		GlueMinerParameters parameters;
		
		GlueMinerWizardStep glueMinerWizardStep = new GlueMinerWizardStep(log);
		List<ProMWizardStep<GlueMinerParameters>> wizStepList = new ArrayList<>();
		wizStepList.add(glueMinerWizardStep);
		ListWizard<GlueMinerParameters> listWizard = new ListWizard<>(wizStepList);
		
		parameters = ProMWizardDisplay.show(context, listWizard, new GlueMinerParameters());
		
		if (parameters != null) {
			GlueMiner miner = new GlueMiner(context, log, parameters);
			miner.mineEntireProcess();
			
			Petrinet net = miner.getNet();
			Marking initialMarking = miner.getInitialMarking();
			Marking finalMarking = miner.getFinalMarking();
			XEventClassifier classifier = miner.getClassifier();
			TransEvClassMapping mapping = miner.getMapping();
			
			context.getConnectionManager().addConnection(new InitialMarkingConnection(net, initialMarking));
			context.getConnectionManager().addConnection(new FinalMarkingConnection(net, finalMarking));
			
			EvClassLogPetrinetConnection evClassLogPetriNetConnection;
			evClassLogPetriNetConnection = new EvClassLogPetrinetConnection("", net, log, classifier, mapping);
			context.getConnectionManager().addConnection(evClassLogPetriNetConnection);
			
			ETCResults conformanceResults = miner.getEtcResults();
			long nonFitTraces = conformanceResults.getnNonFitTraces();
			double fitness = 0.0;
			if (log.size() > 0) {
				fitness = 1.0 - (double)nonFitTraces/(double)log.size();
			}
			
			System.out.println("ETC non fit traces: "+nonFitTraces+" out of "+log.size()+" (fitness: "+fitness+")");
			System.out.println("ETC precision: "+conformanceResults.getEtcp());
			
			ret = new Object[] { net, initialMarking, finalMarking, mapping };
		}
		else {
			context.getFutureResult(0).cancel(true);
			ret = new Object[] { null, null, null, null };
		}
		return ret;
	}
}
