package org.processmining.glueminer.plugins;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginQuality;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.wizard.ListWizard;
import org.processmining.framework.util.ui.wizard.ProMWizardDisplay;
import org.processmining.framework.util.ui.wizard.ProMWizardStep;
import org.processmining.glueminer.algorithms.LogGenerator;
import org.processmining.glueminer.dialogs.LogGeneratorPluginWizardStep;
import org.processmining.glueminer.models.GeneratedPetrinet;
import org.processmining.glueminer.parameters.LogGeneratorParameters;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

@Plugin(name = "Generate log from Petri Net (complex)", level = PluginLevel.NightlyBuild, parameterLabels = { "Generated Petri Net"}, returnLabels = { "Log", "Petrinet", "Initial Marking", "Final Marking", "TransEvClassMapping" }, returnTypes = { XLog.class, Petrinet.class, Marking.class, Marking.class, TransEvClassMapping.class }, help = "No help", quality = PluginQuality.VeryGood, categories = {
		PluginCategory.Discovery })
public class LogGeneratorPluginComplex {
	@UITopiaVariant(affiliation = "PADS", author = "Alessandro Berti", email = "a.berti@pads.rwth-aachen.de")
	@PluginVariant(requiredParameterLabels = { 0 })
	public static Object[] getLogFromGeneratedPetrinet(UIPluginContext context, GeneratedPetrinet gnet) {
		LogGeneratorParameters parameters;
		
		LogGeneratorPluginWizardStep wizardStep = new LogGeneratorPluginWizardStep(gnet.net);
		List<ProMWizardStep<LogGeneratorParameters>> wizStepList = new ArrayList<>();
		wizStepList.add(wizardStep);
		ListWizard<LogGeneratorParameters> listWizard = new ListWizard<>(wizStepList);
		
		parameters = ProMWizardDisplay.show(context, listWizard, new LogGeneratorParameters());
		
		if (parameters != null) {
			LogGenerator logGenerator = new LogGenerator(gnet, parameters);
			
			XLog returnedLog = logGenerator.generateXLog();
			Petrinet returnedNet = gnet.net;
			Marking initialMarking = gnet.initialMarking;
			Marking finalMarking = gnet.finalMarking;
			XEventClassifier classifier = XLogInfoImpl.NAME_CLASSIFIER;
			XEventClass dummyEvClass = new XEventClass("DUMMYDUMMYDUMMY", 999999);
			TransEvClassMapping mapping = new TransEvClassMapping(classifier, dummyEvClass);
			
			XLogInfo summary = XLogInfoFactory.createLogInfo(returnedLog, classifier);
			Collection<XEventClass> logClasses = summary.getEventClasses().getClasses();
			for (XEventClass logClass : logClasses) {
				String activity = logClass.getId();
				Transition transition = gnet.activitiesTransitionCorrespondence.get(activity);
				mapping.put(transition, logClass);
			}
			
			context.getConnectionManager().addConnection(new InitialMarkingConnection(returnedNet, initialMarking));
			context.getConnectionManager().addConnection(new FinalMarkingConnection(returnedNet, finalMarking));
			
			EvClassLogPetrinetConnection evClassLogPetriNetConnection;
			evClassLogPetriNetConnection = new EvClassLogPetrinetConnection("", returnedNet, returnedLog, classifier, mapping);
			context.getConnectionManager().addConnection(evClassLogPetriNetConnection);
			
			return new Object[] { returnedLog, returnedNet, initialMarking, finalMarking, mapping };
		}
		else {
			return new Object[] { null, null, null, null, null };
		}
	}
}
